# @generated

# This file was generated by running `python -m dagster_cloud.pex.grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output


# ruff: noqa: SLF001

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: multi_pex_api.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x13multi_pex_api.proto\x12\rmulti_pex_api"\x07\n\x05\x45mpty"8\n\x16\x43reatePexServerRequest\x12\x1e\n\x16\x63reate_pex_server_args\x18\x01'
    b' \x01(\t":\n\x14\x43reatePexServerReply\x12"\n\x1a\x63reate_pex_server_response\x18\x01'
    b' \x01(\t"4\n\x14GetPexServersRequest\x12\x1c\n\x14get_pex_servers_args\x18\x01'
    b' \x01(\t"6\n\x12GetPexServersReply\x12 \n\x18get_pex_servers_response\x18\x01'
    b' \x01(\t"<\n\x18ShutdownPexServerRequest\x12 \n\x18shutdown_pex_server_args\x18\x01'
    b' \x01(\t">\n\x16ShutdownPexServerReply\x12$\n\x1cshutdown_pex_server_response\x18\x01'
    b' \x01(\t"\x1b\n\x0bPingRequest\x12\x0c\n\x04\x65\x63ho\x18\x01'
    b' \x01(\t"\x19\n\tPingReply\x12\x0c\n\x04\x65\x63ho\x18\x01'
    b' \x01(\t2\xf0\x02\n\x0bMultiPexApi\x12_\n\x0f\x43reatePexServer\x12%.multi_pex_api.CreatePexServerRequest\x1a#.multi_pex_api.CreatePexServerReply"\x00\x12Y\n\rGetPexServers\x12#.multi_pex_api.GetPexServersRequest\x1a!.multi_pex_api.GetPexServersReply"\x00\x12\x65\n\x11ShutdownPexServer\x12\'.multi_pex_api.ShutdownPexServerRequest\x1a%.multi_pex_api.ShutdownPexServerReply"\x00\x12>\n\x04Ping\x12\x1a.multi_pex_api.PingRequest\x1a\x18.multi_pex_api.PingReply"\x00\x62\x06proto3'
)


_EMPTY = DESCRIPTOR.message_types_by_name["Empty"]
_CREATEPEXSERVERREQUEST = DESCRIPTOR.message_types_by_name["CreatePexServerRequest"]
_CREATEPEXSERVERREPLY = DESCRIPTOR.message_types_by_name["CreatePexServerReply"]
_GETPEXSERVERSREQUEST = DESCRIPTOR.message_types_by_name["GetPexServersRequest"]
_GETPEXSERVERSREPLY = DESCRIPTOR.message_types_by_name["GetPexServersReply"]
_SHUTDOWNPEXSERVERREQUEST = DESCRIPTOR.message_types_by_name["ShutdownPexServerRequest"]
_SHUTDOWNPEXSERVERREPLY = DESCRIPTOR.message_types_by_name["ShutdownPexServerReply"]
_PINGREQUEST = DESCRIPTOR.message_types_by_name["PingRequest"]
_PINGREPLY = DESCRIPTOR.message_types_by_name["PingReply"]
Empty = _reflection.GeneratedProtocolMessageType(
    "Empty",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.Empty)
    },
)
_sym_db.RegisterMessage(Empty)

CreatePexServerRequest = _reflection.GeneratedProtocolMessageType(
    "CreatePexServerRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEPEXSERVERREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.CreatePexServerRequest)
    },
)
_sym_db.RegisterMessage(CreatePexServerRequest)

CreatePexServerReply = _reflection.GeneratedProtocolMessageType(
    "CreatePexServerReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEPEXSERVERREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.CreatePexServerReply)
    },
)
_sym_db.RegisterMessage(CreatePexServerReply)

GetPexServersRequest = _reflection.GeneratedProtocolMessageType(
    "GetPexServersRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPEXSERVERSREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.GetPexServersRequest)
    },
)
_sym_db.RegisterMessage(GetPexServersRequest)

GetPexServersReply = _reflection.GeneratedProtocolMessageType(
    "GetPexServersReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPEXSERVERSREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.GetPexServersReply)
    },
)
_sym_db.RegisterMessage(GetPexServersReply)

ShutdownPexServerRequest = _reflection.GeneratedProtocolMessageType(
    "ShutdownPexServerRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SHUTDOWNPEXSERVERREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.ShutdownPexServerRequest)
    },
)
_sym_db.RegisterMessage(ShutdownPexServerRequest)

ShutdownPexServerReply = _reflection.GeneratedProtocolMessageType(
    "ShutdownPexServerReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _SHUTDOWNPEXSERVERREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.ShutdownPexServerReply)
    },
)
_sym_db.RegisterMessage(ShutdownPexServerReply)

PingRequest = _reflection.GeneratedProtocolMessageType(
    "PingRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _PINGREQUEST,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.PingRequest)
    },
)
_sym_db.RegisterMessage(PingRequest)

PingReply = _reflection.GeneratedProtocolMessageType(
    "PingReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _PINGREPLY,
        "__module__": "multi_pex_api_pb2",
        # @@protoc_insertion_point(class_scope:multi_pex_api.PingReply)
    },
)
_sym_db.RegisterMessage(PingReply)

_MULTIPEXAPI = DESCRIPTOR.services_by_name["MultiPexApi"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _EMPTY._serialized_start = 38
    _EMPTY._serialized_end = 45
    _CREATEPEXSERVERREQUEST._serialized_start = 47
    _CREATEPEXSERVERREQUEST._serialized_end = 103
    _CREATEPEXSERVERREPLY._serialized_start = 105
    _CREATEPEXSERVERREPLY._serialized_end = 163
    _GETPEXSERVERSREQUEST._serialized_start = 165
    _GETPEXSERVERSREQUEST._serialized_end = 217
    _GETPEXSERVERSREPLY._serialized_start = 219
    _GETPEXSERVERSREPLY._serialized_end = 273
    _SHUTDOWNPEXSERVERREQUEST._serialized_start = 275
    _SHUTDOWNPEXSERVERREQUEST._serialized_end = 335
    _SHUTDOWNPEXSERVERREPLY._serialized_start = 337
    _SHUTDOWNPEXSERVERREPLY._serialized_end = 399
    _PINGREQUEST._serialized_start = 401
    _PINGREQUEST._serialized_end = 428
    _PINGREPLY._serialized_start = 430
    _PINGREPLY._serialized_end = 455
    _MULTIPEXAPI._serialized_start = 458
    _MULTIPEXAPI._serialized_end = 826
# @@protoc_insertion_point(module_scope)
